#include "simulator.h"
static void SDQ_compute(int *edge,int *field,int *id,int *lnd,double *result);		// function to compute landscape surface metrics

void *simulator(void *thrdata) {		// Landscape simulator and analyzer, this function is run from within separate threads
	THREADDATA *tdt=(THREADDATA*)thrdata;
	int *pmanintervals=tdt->manint;
	OWNER *owli=tdt->ownerlist;
/*
	Landscape simulator and analyzer
	
	landscape: initial landscape, an object of class hexgrid with two fields for each cell: owner and age (time since last management)
	pred: predicted values of models, one value per year. Currently only one model is supported
	rgama: look-up table with random values from a gamma distribution (see script)
	manintervals: management intervals, one value for each owner group. First value corresponds to group 1.
	nyears: number of simulation years to run
	owners: list of owners, respective owned cells (their indexes) and owner group starting in 1 (see script). Group 0 is reserved for the no-management case
	ageintervals: intervals of age classes whose area is to be recorded (the minimum area)
	startyear: only compute objectives for this year onwards (this is to allow for landscape adaptation to the new regime)
	trace: for tracing purposes only. Used to analyse in detail the changes in objectives during simulation.
*/
	// get the pointers
	int i,j;
	int *tsd=tdt->tsd;	// pointer to cell age matrix
	OWNERCELLS *ownercells=tdt->workinglandscape->ownercells;
	int *edge=INTEGER_POINTER(VECTOR_ELT(oriland,2));					// p to cell neighboring codes
	int *id=INTEGER_POINTER(VECTOR_ELT(oriland,0));					// p to cell IDs (only useful when NAs are present, e.g. for non-rectangular landscapes)
	int *pgama=INTEGER_POINTER(rgama),*ptrace=0,lentrace=0;
	int *ppred=INTEGER_POINTER(pred);
	int *dim=INTEGER_POINTER(GET_DIM(VECTOR_ELT(VECTOR_ELT(oriland,1),1)));	// landscape dimensions
	int ncell=LENGTH(VECTOR_ELT(VECTOR_ELT(oriland,1),0));		// nr. of cells in the landscape
	int startingyear=INTEGER_POINTER(startyear)[0];
	int nmods=INTEGER_POINTER(GET_DIM(pred))[1];				// nr. of models to use (currently only one is supported)
//	int nowners=LENGTH(owners);								// nr. of owners	
	short ngama=INTEGER_POINTER(GET_DIM(rgama))[0];
	int *intervalostsd=INTEGER_POINTER(intervals);
	short nintervalostsd=GET_LENGTH(intervals)-1;
	short max_ano=INTEGER_POINTER(GET_DIM(pred))[0]-1;		// maximum age for which there are model predictions (e.g. biomass values)
	short NOBJ=4+nintervalostsd;
	int year,tmp,leng=(dim[0]*2-1)*(dim[1]*2-1),tmpinter[nintervalostsd];
	double var[2];
		
	int *carga=(int*)malloc(sizeof(int)*dim[0]*dim[1]);		// to store biomasses per cell
	int *edgework=(int*)malloc(sizeof(int)*leng);			// temporary buffer for edge contrast and skewness computations
	
	memset(carga,0,sizeof(int)*dim[0]*dim[1]);

/*	PROTECT(obj=NEW_NUMERIC(NOBJ));	// to store objectives
	double *pobj=NUMERIC_POINTER(obj),*ptrobj=0,objinc[NOBJ];*/
	double *ptrobj=0,objinc[NOBJ];
	double *pobj=tdt->pobj;
	//memset(pobj,0,sizeof(double)*NOBJ);
	
	int protcount=0;
	double minmaxobj[NOBJ];
	int *ptrtsd,*ptrcarga;		// for tracing purposes
	char tmpbuf[120];

	if(tdt->tracelist) {
		lentrace=LENGTH(trace);
		ptrace=INTEGER_POINTER(trace);
	}
	
	// initialize owner list
	for(i=0;i<tdt->workinglandscape->nowners;i++) {
		if(ownercells[i].ownedcells!=NULL) {			// number of owned cells >0
			if(owli[i].group<0) {						// group -1 does not manage ever
				owli[i].interval=0;
				owli[i].nextmanyear=-1;
			} else {
				if(pmanintervals[owli[i].group]==NA_INTEGER) {	// if management interval for this group is NA, owners never manage
					owli[i].interval=0;
					owli[i].nextmanyear=-1;
				} else {	// otherwise, compute the first year of management based on random values from a gamma distribution with mean equal to desired interval
					owli[i].interval=pmanintervals[owli[i].group];
					// next man.year is calculated taking the last man.date (prior to simulation) as the reference
					owli[i].nextmanyear=pgama[ngama*(owli[i].interval-1) + (int)((float)rand()/RAND_MAX*ngama)] - tsd[ownercells[i].ownedcells[0]];
					// if the owner should have managed already, then assume he manages in the first year
					if(owli[i].nextmanyear<0) owli[i].nextmanyear=0;		// if he should have managed already according to this regime, then manage now (in the first year of simulation)
					// NOTE: the above line causes synchronization among many owners when intervals are low. It's the cause of the cyclic patterns in FIG.5
				}
			}
		}
	}
	
	// all objectives except the first (0) are the extremes (min or max), not the sum.
	// array minmaxobj will store the extremes.
	minmaxobj[1]=-100000000;		// total biomass
	minmaxobj[2]=100000000;		// edge contrast variance
	minmaxobj[3]=100000000;		// skewness
	for(i=0;i<nintervalostsd;i++) minmaxobj[4+i]=100000000;		// area within each age class
	
	for(year=0;year<nyears;year++) {	// main simulation loop
		memset(objinc,0,NOBJ*sizeof(double));		// this variable is for tracing purposes only, it's cleared every iteration (year)
		
		ptrcarga=NULL;ptrtsd=NULL;
			
		// for each owner, check if this year is a management year
		for(j=0;j<tdt->workinglandscape->nowners;j++) { 
			if(year==owli[j].nextmanyear) {	// it's a management year for this owner. MANAGE!
				for(i=0;i<ownercells[j].ncells;i++) {
					tsd[ownercells[j].ownedcells[i]]=0;		// set all cells owned to 0 age
					pobj[0]++;					// OBJECTIVE: count managed area (sum)
					objinc[0]++;				// for tracing
				}
				// now re-assign the next management year, based on random values from a gamma distribution with mean equal to the desired management interval 
				owli[j].nextmanyear+=pgama[ngama*(owli[j].interval-1) + (int)((float)rand()/RAND_MAX*ngama)];
			}
		}
		
		// is it a trace year? should we save the values? then get pointers to output matrices
		for(i=0;i<lentrace;i++) {
			if(ptrace[i]==year) {	// yes, prepare pointers to save values
				if(savelandscapes) {
					//ptrcarga=INTEGER_POINTER(VECTOR_ELT(tdt->tracelist,0))+i*dim[0]*dim[1];
					ptrtsd=INTEGER_POINTER(VECTOR_ELT(tdt->tracelist,2))+i*dim[0]*dim[1];
				}
				ptrobj=NUMERIC_POINTER(VECTOR_ELT(tdt->tracelist,1))+i;
				break;
			}
		}
		
		// compute objectives only after startingyear to allow for adaptation to the new regime (except the first, which is computed for the whole period)
#ifdef EVENYEARS
		if((year>=startingyear || ptrobj) && (year % 2)) {
#else
		if(year>=startingyear || ptrobj) {
#endif
			memset(tmpinter,0,nintervalostsd*sizeof(int));	// for counting the nr. of cells with ages within the given intervals
			for(i=0;i<ncell;i++) {
#ifdef NONAS
	#ifdef ONEMODELONLY
				if(tsd[i]>max_ano)	// if cell age is outside prediction range (0-70 yr)... 
					tmp=ppred[max_ano];	// assume biomass is equal to 70 yr age
				else	// otherwise just take biomass from model
					tmp=ppred[tsd[i]];
				carga[i]=tmp;
				pobj[1]+=tmp;			// OBJECTIVE: total biomass in the landscape (later on we'll take the maximum)
				objinc[1]+=tmp;
	#else
				for(j=0;j<nmods;j++) {	// for each model NOTE: only one is supported...
					// compute predicted biomass
					if(tsd[i]>max_ano)	// if cell age is outside prediction range (0-70 yr)... 
						tmp=ppred[max_ano+j*(max_ano+1)];	// assume biomass is equal to 70 yr age
					else	// otherwise just take biomass from model
						tmp=ppred[tsd[i]+j*(max_ano+1)];
					carga[i]=tmp;
					pobj[1]+=tmp;			// OBJECTIVE: total biomass in the landscape (later on we'll take the maximum)
					objinc[1]+=tmp;
				}
	#endif				
				for(j=0;j<nintervalostsd;j++) {		// count the nr. of cells within the given intervals
					if(tsd[i]>=intervalostsd[j] && tsd[i]<intervalostsd[j+1]) tmpinter[j]++;
				}
#else
				if(tsd[i]!=NA_INTEGER) {
	#ifdef ONEMODELONLY
					if(tsd[i]>max_ano)	// if cell age is outside prediction range (0-70 yr)... 
						tmp=ppred[max_ano];	// assume biomass is equal to 70 yr age
					else	// otherwise just take biomass from model
						tmp=ppred[tsd[i]];
					carga[i]=tmp;
					pobj[1]+=tmp;			// OBJECTIVE: total biomass in the landscape (later on we'll take the maximum)
					objinc[1]+=tmp;
	#else				
					for(j=0;j<nmods;j++) {	// for each model NOTE: only one is supported...
						// compute predicted biomass
						if(tsd[i]>max_ano)	// if cell age is outside prediction range (0-70 yr)... 
							tmp=ppred[max_ano+j*(max_ano+1)];	// assume biomass is equal to 70 yr age
						else	// otherwise just take biomass from model
							tmp=ppred[tsd[i]+j*(max_ano+1)];
						carga[i]=tmp;
						pobj[1]+=tmp;			// OBJECTIVE: total biomass in the landscape (later on we'll take the maximum)
						objinc[1]+=tmp;
					}
	#endif			
					for(j=0;j<nintervalostsd;j++) {		// count the nr. of cells within the given age class intervals
						if(tsd[i]>=intervalostsd[j] && tsd[i]<intervalostsd[j+1]) tmpinter[j]++;
					}
				} else {
					carga[i]=-1;
				}
#endif
			}
			
			// compute edge contrast variance and skewness based on predicted biomass values per cell
			memcpy(edgework,edge,leng*sizeof(int));	
			SDQ_compute(edgework,carga,id,dim,var);
			// var[0]: edge contrast variance
			// var[1]: biomass skewness

			if(pobj[1]>minmaxobj[1]) minmaxobj[1]=pobj[1];		// OBJECTIVE: take the maximum total biomass attained during simulation
			if(var[0]<minmaxobj[2]) minmaxobj[2]=var[0];		// OBJECTIVE: take the minimum edge contrast variance attained during simulation
			if(var[1]<minmaxobj[3]) minmaxobj[3]=var[1];		// OBJECTIVE: the same but for skewness
			for(j=0;j<nintervalostsd;j++) {						// OBJECTIVES: take the minimum area within each age class attained in one year during simulation
				if(tmpinter[j]<minmaxobj[4+j]) minmaxobj[4+j]=tmpinter[j];	
			}
			pobj[1]=0;
			
			objinc[2]=-var[0];
			objinc[3]=-var[1];
			if(ptrobj && savelandscapes) {	// this is a trace year, so save values
//				memcpy(ptrcarga,carga,sizeof(int)*dim[0]*dim[1]);
				memcpy(ptrtsd,tsd,sizeof(int)*dim[0]*dim[1]);
			}
		}
		
		if(ptrobj) {	// this is a trace year, so save values
			*ptrobj=objinc[0];
			*(ptrobj+lentrace)=objinc[1];
			*(ptrobj+lentrace*2)=objinc[2];
			*(ptrobj+lentrace*3)=objinc[3];
			for(j=0;j<nintervalostsd;j++) *(ptrobj+lentrace*(4+j))=(tmpinter[j]==NA_INTEGER ? NA_REAL : tmpinter[j]);
		}
#ifdef NONAS
		for(i=0;i<ncell;i++) {tsd[i]++;}		// advance one year
#else
		for(i=0;i<ncell;i++) {if(tsd[i]!=NA_INTEGER) tsd[i]++;}		// advance one year
#endif
	}	// END main simulation loop

	// prepare objectives for output (pobj[0] is ready already, cause it's the sum)
	pobj[1]=minmaxobj[1];
	pobj[2]=minmaxobj[2];
	pobj[3]=minmaxobj[3];
	for(j=0;j<nintervalostsd;j++) pobj[4+j]=(minmaxobj[4+j]==100000000 ? NA_REAL : minmaxobj[4+j]);
	
	free(carga);
	free(edgework);
	tdt->ready=1;
	tdt->returned=1;
	return 0;
}

void constraintFunction(double **par,double **out,SEXP objfuncpars) {
	int npars=INTEGER_POINTER(getListElement(objfuncpars,"npars"))[0];
	if(npars<=3) *out[0]=1;
	int nmanint=(npars+1)/2,npropor=(npars-1)/2,i;
	double soma=0;
	// this ensures that all the n.groups have at least 1% of owners assigned
	for(i=0;i<npropor;i++) soma+=*par[i+nmanint];		// par[i+nmanint] are the proportion of owners in each group
	*out[0]=0.99-soma;		// negative when our requirement above does not verify
}

static void SDQ_compute(int *edge,int *field,int *id,int *lnd,double *result) {		// function to compute landscape surface metrics
	// actually does the computations for both indices
	register int x,y,i,j;
	short nw1=-lnd[0]*2,w1=-lnd[0]*2+1,n1=-1,ne1=lnd[0]*2-2;
	short nw2=-lnd[0]*4,w2=-lnd[0]*4+2,n2=-2,ne2=lnd[0]*4-4;
	unsigned long long ss=0,s=0,s3=0,s1=0,s2=0;
	int *tmp,count=0,countcells=0,jinc=(lnd[0]<<1)-2;
	double tmp1,tmp2,tmp3;

	for(x=0,i=0,j=0;x<lnd[1];x++,j+=jinc) { // copy values of biomass to the central cells of EDGE, and at the same time computes needed sums for skewness computation
		for(y=0;y<lnd[0];y++,i++,j+=2) {
#ifdef NONAS
			edge[j]=field[i];
			s1+=field[i];
			s2+=(unsigned long long)field[i]*field[i];
			s3+=(unsigned long long)field[i]*field[i]*field[i];
			countcells++;
#else
			if(id[i]!=NA_INTEGER) {
				edge[j]=field[i];
				s1+=field[i];
				s2+=(unsigned long long)field[i]*field[i];
				s3+=(unsigned long long)field[i]*field[i]*field[i];
				countcells++;
			} else edge[j]=NA_INTEGER;
#endif
		}
	}

	for(x=0,i=0,j=0;x<lnd[1];x++,j+=jinc) {	// computes edge contrast for all edges
		for(y=0;y<lnd[0];y++,i++,j+=2) {
#ifdef NONAS
			if(x>0) {
				tmp=&edge[j+w1];*tmp=abs(edge[j]-edge[j+w2]);ss+=(*tmp)*(*tmp);s+=*tmp;count++;
				if(y>0) {
					if(edge[j+nw1]==-4) {
						tmp=&edge[j+nw1];*tmp=abs(edge[j]-edge[j+nw2]);ss+=(*tmp)*(*tmp);s+=*tmp;count++;
					}
				}
			}
			if(y>0) {
				tmp=&edge[j+n1];*tmp=abs(edge[j]-edge[j+n2]);ss+=(*tmp)*(*tmp);s+=*tmp;count++;
				if(x<lnd[1]-1) {
					if(edge[j+ne1]==-3) {
						tmp=&edge[j+ne1];*tmp=abs(edge[j]-edge[j+ne2]);ss+=(*tmp)*(*tmp);s+=*tmp;count++;
					}
				}
			}

#else		
			if(id[i]!=NA_INTEGER) {
				if(x>0) {
					if(edge[j+w2]!=NA_INTEGER) {tmp=&edge[j+w1];*tmp=abs(edge[j]-edge[j+w2]);ss+=(*tmp)*(*tmp);s+=*tmp;count++;} 
					if(y>0) {
						if(edge[j+nw1]==-4) {
							if(edge[j+nw2]!=NA_INTEGER) {tmp=&edge[j+nw1];*tmp=abs(edge[j]-edge[j+nw2]);ss+=(*tmp)*(*tmp);s+=*tmp;count++;}
						}
					}
				}
				if(y>0) {
					if(edge[j+n2]!=NA_INTEGER) {tmp=&edge[j+n1];*tmp=abs(edge[j]-edge[j+n2]);ss+=(*tmp)*(*tmp);s+=*tmp;count++;}
					if(x<lnd[1]-1) {
						if(edge[j+ne1]==-3) {
							if(edge[j+ne2]!=NA_INTEGER) {tmp=&edge[j+ne1];*tmp=abs(edge[j]-edge[j+ne2]);ss+=(*tmp)*(*tmp);s+=*tmp;count++;}
						}
					}
				}
			}
#endif
		}
	}
	tmp1=(double)s/count;
	result[0]=(((double)ss/count)-(tmp1*tmp1));	
	tmp1=(double)s1/countcells;
	tmp2=(double)s2/countcells;
	tmp3=sqrt(tmp2-tmp1*tmp1);
	if(tmp3==0) {	// Undefined skewness
		//Rprintf("Undefined skewness!\n");
		result[1]=0;
	} else
		result[1]=(((double)s3/countcells)-3*tmp1*tmp2+2*tmp1*tmp1*tmp1)/(tmp3*tmp3*tmp3);
	return;
}

SEXP SDQ(SEXP hexg,SEXP field) {
	// computes contrast variance and skewness of a variable in an hexgrid object
	// field is the 1-based index of the data in hexg$data
	// just a wrapper for SDQ_compute	
	int *edge=INTEGER_POINTER(VECTOR_ELT(hexg,2));
	int *id=INTEGER_POINTER(VECTOR_ELT(hexg,0));
	int *campo=INTEGER_POINTER(VECTOR_ELT(VECTOR_ELT(hexg,1),INTEGER_POINTER(field)[0]-1));
	int *dim=INTEGER_POINTER(GET_DIM(VECTOR_ELT(hexg,0)));	
	int len=(dim[0]*2-1)*(dim[1]*2-1);
	int *tmp=(int*)malloc(sizeof(int)*len);
	double *pretval;
	SEXP retval;
	PROTECT(retval=NEW_NUMERIC(2));
	pretval=NUMERIC_POINTER(retval);

	memcpy(tmp,edge,len*sizeof(int));	
	SDQ_compute(tmp,campo,id,dim,pretval);
	free(tmp);
	UNPROTECT(1);
		SEXP names;	
		PROTECT(names=NEW_CHARACTER(2));
		SET_STRING_ELT(names,0,mkChar("Contrast variance"));
		SET_STRING_ELT(names,1,mkChar("Skewness"));
		SET_NAMES(retval,names);
		UNPROTECT(1);
	return(retval);
}


