output="tif"	# output plots either as tif or eps

if(!all(file.exists(c("helper_functions.r","plotting_functions.r")))) stop("Some script or data files not found.\n\nYou must set R working directory to the folder where all the files are unpacked, with setwd() or clicking menu File | Change dir...")
source("plotting_functions.r")
source("helper_functions.r")

# create a landscape grid 18x18 km composed of 1ha hexagonal cells
landscape=newHexgrid(c(0, 18000),c(0, 18000),a=10000)
landscape$data$ownerID=landscape$id	# just a placeholder matrix for writing owner IDs of each cell later
landscape$data$tsd=landscape$id		# just a placeholder matrix for time since management of each cell to be set later

# sample data for the biomass accumulation curve as an integer look-up table
# data comes from leaf+fine material biomass Generalized Additive Model predictions for each year after management, see paper
# Porto, M., Correia, O., Beja, P., 2013. Modelling fuel succession in Mediterranean cork oak forests along a 70-year chronosequence. Forest Ecology and Management 302, 221–230.
biomass.curve=matrix(as.integer(c(428, 763, 1063, 1337, 1588, 1818, 2029, 2225, 2406, 2573, 2730, 2875, 3012, 3139, 3258, 3371, 3477, 3577, 3671, 3761, 3846, 3927, 4005, 4079, 4150, 4217, 4282, 4345, 4405, 4463, 4519, 4573, 4625, 4675, 4724, 4772, 4818, 4862, 4906, 4948, 4990, 5030, 5069, 5108, 5146, 5183, 5219, 5254, 5289, 5323, 5357, 5390, 5422, 5454, 5485, 5516, 5547, 5577, 5607, 5636, 5665, 5694, 5722, 5750, 5777, 5804, 5831, 5858, 5885, 5911, 5936)),nc=1)

# create a look-up table used in simulations for randomly choosing a management interval based on the desired interval that is asked to the owner
# "uncertainty" is the variance of the distribution at the highest interval ("maxInterval");
# "fac" determines how much the variance increases with the mean, and varies from 0 to 1:
#		fac=0 means that there is no correlation between mean and variance, being variance=uncertainty (approx.) in all intervals;
#		fac=1 yields the same distribution used in the paper, with variance=uncertainty in the highest interval (100), zero in the lowest interval (1), following a quadratic curve in between;
#			intermediate values of fac lie between these two extremes.
maxInterval=100
rgama=get.interval.gamma(uncertainty=100,fac=1,maxInterval=maxInterval)

#********************************* set parameters for optimization and simulation
otherpars=list(
# Parameters for optimization
	pop=800L							# population of genetic algorithm (set to 800 in the paper)
	,gener=as.integer(c(10, 20, 50, 100, 150, 200, 250, 300, 350, 400))		# number of generations of genetic algorithm (set to 400 in the paper)
# Parameters for the objectives
	,intervals=as.integer(c(30,60,9999))	# Biodiversity objectives: age class intervals that will be required not fall below given proportions (see next parameter) (must be integer)
	,minareaTSD=c(0.10,0.10)			# Biodiversity objectives: proportion of area of each age class that is required as a minimum to keep during all the simulation period.
							# In this example, a minimum of 10% of the landscape must be maintained within [30,60[ yr; and 10% greater than 60 yr.
	,minInterval=3						# minimum and maximum values that the management interval decision variables can take during optimization			
	,maxInterval=as.integer(maxInterval)# note that when an interval (during optimization) is greater than maxInterval-1, it is taken as meaning "no-management"
# Parameters for simulation
	,n.years=100L						# number of years to run simulation (must be integer)
	,startyear=20L						# simulation year when fire risk and biodiversity objectives start to be computed (to allow for landscape "adaptation" to the new regime)
	,models=biomass.curve				# model to be used by the simulator for biomass predictions
	,landscape=landscape				# landscape grid (only structure matters, data will be set in the C routines during optimization)
	,n.landscape.repetitions=24L		# in how many random landscapes each solution is evaluated during optimization
	,n.repeats=1L						# number of reevaluations of the same solution in the same landscape, the objective values are taken as the maximum of all evaluations
	,rgama=rgama						# look-up table of gamma values to introduce random deviations from desired management intervals
	,total.area=as.integer(sum(!is.na(landscape$id)))	# total landscape area (to compute objective values during optimization)	
# How many parallel threads
# NOTE: faster calculations when n.landscape.repetitions*n.repeats is a multiple of n.threads.
	,n.threads=12L						# number of threads to use is parallel (hopefully, one in each processor core)	
);

otherpars$obj.names=c(			# names of optimization objectives
	"Average managed area per year (% from total)"
	,"Maximum biomass per hectare per year"
	,"Minimum edge contrast variance"
	,"Minimum biomass skewness"
	,paste("Minimum area between",otherpars$intervals[1:(length(otherpars$intervals)-2)],"and",otherpars$intervals[2:(length(otherpars$intervals)-1)],"yr (%)")
	,paste("Minimum area >",otherpars$intervals[length(otherpars$intervals)-1],"yr (%)")
)
otherpars$obj.shortnames=c(		# short names of optimization objectives
	"Av man area/yr"
	,"Max bio/ha.yr"
	,"Min edge contr var"
	,"Min bio skew"
	,paste("Min area ",otherpars$intervals[1:(length(otherpars$intervals)-2)],"-",otherpars$intervals[2:(length(otherpars$intervals)-1)],"yr")
	,paste("Min area >",otherpars$intervals[length(otherpars$intervals)-1],"yr")
)
otherpars$partial.obj.names=c(	# names of yearly optimization objectives (values passed to the algorithm are minima and maxima of these)
	"Managed area (% from total)"
	,"Biomass per hectare"
	,"Edge contrast variance"
	,"Biomass skewness"
	,paste("Area between",otherpars$intervals[1:(length(otherpars$intervals)-2)],"and",otherpars$intervals[2:(length(otherpars$intervals)-1)],"yr (%)")
	,paste("Area >",otherpars$intervals[length(otherpars$intervals)-1],"yr (%)")
)

NOBJ=4+length(otherpars$minareaTSD)	# number of optimization objectives
objective.fn=function(par,otherpars) {}		# this is a dummy objective function, the real objective function is hardcoded in the C code.

############################################################################################################

cat("What do you want?\n")
switch(menu(c("Explore results from paper","Make random landscapes","Run optimization analysed in the paper (takes days)","Influence of a priori assumptions","Simulate solutions along time","Evaluate convergence","Plot implementation uncertainty distributions")),
{
	cat("Which run?\n")
	files=dir("Results/",pattern="optim_")
	file=paste("Results/",files[menu(files)],sep="")
	load(file)
	if(!exists("results")) results=res
	if(length(results)>1) {
		cat("There are",length(results),"runs in this file. Which one do you want?\n")
		if(is.null(names(results)))
			run=menu(1:length(results))
		else
			run=menu(names(results))
		results=results[[run]]
	} else results=results[[1]]
	results=results[[length(results)]]		# extract the results from the last generation step only

	limits=setLimits(results)				# compute restrictions for the restriction scenarios (using percentiles)
	
	limits=switch(menu(c("No restrictions"
		,"Biodiversity priority: minimum 9.9% area in each age class"
		,"Fire risk priority: fire risk objectives within the extreme 25%"
		,"Cost efficiency 5%: managed area between 5%+-0.375%"
		,"Cost efficiency 10%: managed area between 10%+-0.375%"
		,"\"Compromise solutions\""
		,"All the above (Figure 2)"
		)),limits[1],limits[2],limits[3],limits[4],limits[5],limits[6],limits[1:6]
	)
	
	switch(menu(c("Level plots of objectives","Pairwise objective scatterplots","Plot solutions in detail","Plot solution ordination")),{
		if(length(limits)>1) stop("Only one constraint at a time is supported in the level plots.")
		openDevice(output,name="levelplots",wid=2*3.5,hei=3*3.5)
	   	solutions=visualize(results,disp="o",limits=limits[[1]],color=is.null(limits[[1]]),dispminimum=NA,ranges=list(NULL,NULL,NULL,c(-5,5),NULL,NULL))
	   	dev.off()
	},{
		if(length(limits)>1) stop("Only one constraint at a time is supported in the pairwise scatterplots.")
		openDevice(output,name="objectivematrix",wid=3*3.5,hei=5*3.5)
		plotObjectiveMatrix(results,limits=limits[[1]],ranges=list(NULL,NULL,NULL,c(-5,5),NULL,NULL))
		dev.off()
		if(F) {
			for(i in 1:length(results[[1]])) {
			res=lapply(results,"[[",i)		# extract the results from the last generation step only
			plotObjectiveMatrix(res,limits=limits,ranges=list(NULL,NULL,NULL,c(-5,5),NULL,NULL))
			readline()
			}
		}
	},{
		if(length(limits)==6) {
			openDevice(output,name="solutions",wid=2*3.5,hei=3*3.5)
			layout(matrix(1:length(limits),nc=2,byr=T))
		} else {
			openDevice(output,name="solutions",wid=7,hei=7)
			layout(matrix(1:length(limits),nc=1,byr=T))
		}
		for(i in 1:length(limits)) {
			if(is.null(limits[[i]]) && length(limits)==6) {
				solutions=visualize(results,disp="p",limits=limits[[i]],dispminimum=1:10,title=names(limits)[i],letter=ifelse(length(limits)>1,LETTERS[i],""),legend=ifelse(i==1,T,F))
			} else {
				solutions=visualize(results,disp="p",limits=limits[[i]],dispminimum=NA,title=names(limits)[i],legend=ifelse(i==1,T,F),letter=ifelse(length(limits)>1,LETTERS[i],""))
			}
		}
		dev.off()
	},{
		if(length(limits)>1) stop("Only one constraint at a time is supported in the ordination plots.")
		results=transformObjectives(results)
		ssr=subset(results,limits[[1]])
		openDevice(output,name="solutions_ord",wid=2*3.5,hei=2*3.5)
		distances=ordinateSolutions(ssr$subsetted$par)
		dev.off()
	})
},{
	if(!require(mcomp)) stop("You must compile and install the modified package in order to run this option.")
	cat("How many landscapes you want to generate? ")
	nl=as.numeric(readline())
	cat("Enter mean property area (20): ")
	meanp=as.numeric(readline())
	if(is.na(meanp)) meanp=20
	cat("Enter variance of property area (100): ")
	varp=as.numeric(readline())
	if(is.na(varp)) varp=100
	makeLandscapes(meanp,varp,nl,landscape,maxTSD=70,threads=otherpars$n.threads,filename=paste("landscapes_",meanp,"_",varp,".bin",sep=""))
},{
	if(!require(mcomp)) stop("You must compile and install the modified package in order to run this option.")
	if(length(dir("./",pattern=".bin"))==0) stop("You must create the landscape file before, with mean=20, variance=100. Run #1.")
######################### classic optimization runs: same as done in original manuscript
	schedule=list(
		list(
			task="optim",
			uncertainties=c(100),
			increasing.uncertainty=c(1),
			landscape.files="landscapes_20_100.bin",
			n.groups=5
		)
	)
	
	results=processTasks(schedule,pause=T)
	save(results,file=paste("optim_020_100_100_1.rdata",sep=""))	# mean-landscape_var-landscape_uncertainty_factor
},{
	otherpars$n.landscape.repetitions=12L
	message("Changed to 12 landscape evaluations instead of 24.")
	switch(menu(c("Optimize - varying landscape properties (mean and variance), keeping sqrt(var)/mean constant"
			,"Optimize - varying implementation uncertainty"
			,"Optimize - varying correlation between mean and variance of implementation"
			,"Compare solutions across different optimizations (signatures and ordination)"
			,"Compare objectives across different optimizations"
			)),{
		if(!require(mcomp)) stop("You must compile and install the modified package in order to run this option.")
		schedule=list(
			list(
				task="optim",
				uncertainties=100,
				increasing.uncertainty=1,
				landscape.files=c("landscapes_20_100.bin","landscapes_60_900.bin","landscapes_100_2500.bin","landscapes_140_4900.bin","landscapes_200_10000.bin"),
				n.groups=5
			)
		)
		results=processTasks(schedule,pause=T)
		save(results,file=paste("optim_---_---_100_1.rdata",sep=""))		
	},{
		if(!require(mcomp)) stop("You must compile and install the modified package in order to run this option.")
		schedule=list(
			list(
				task="optim",
				uncertainties=c(1,5,10,50,100,500,1000),
				increasing.uncertainty=1,
				landscape.files=c("landscapes_20_100.bin"),
				n.groups=5
			)
		)
		results=processTasks(schedule,pause=T)
		save(results,file=paste("optim_020_100_---_1.rdata",sep=""))	# mean-landscape_var-landscape_uncertainty_factor		
	},{
		if(!require(mcomp)) stop("You must compile and install the modified package in order to run this option.")
		schedule=list(
			list(
				task="optim",
				uncertainties=100,
				increasing.uncertainty=c(0,0.25,0.5,0.75,1),		# the 1 can be borrowed from above
				landscape.files=c("landscapes_20_100.bin"),
				n.groups=5
			)
		)
		results=processTasks(schedule,pause=T)
		save(results,file=paste("optim_020_100_100_--.rdata",sep=""))	# mean-landscape_var-landscape_uncertainty_factor
#############		
	},{
		files=dir("Results/",pattern="optim_")
		file=paste("Results/",files[menu(files)],sep="")
		load(file)
		if(!exists("results")) results=res
		if(length(results)<2) stop("This file has only one optimization run.")
		limits=setLimits(results[[1]][[length(results[[1]])]])[1:6]
		title=attr(results,"title")
		baseindex=attr(results,"baseindex")
		
		# plot "signatures" of optimal solutions
		for(i in 1:6) {		# each restriction scenario
			bl=best.layout(length(results),2)
			openDevice(output,name=paste("solutions",gsub("[-%]","",names(limits)[i]),sep=""),wid=dim(bl)[2]*3.5,hei=dim(bl)[1]*3.5)
			layout(bl)
			print(names(limits)[i])
			for(j in 1:length(results)) {
				limits=setLimits(results[[j]][[length(results[[j]])]])[1:6]
				if(is.null(limits[[i]]))
					visualize(results[[j]][[length(results[[j]])]],disp="p",dispminimum=1:10,limits=limits[[i]],title=names(results)[j],letter=LETTERS[j],legend=ifelse(j==1,T,F))
				else
					visualize(results[[j]][[length(results[[j]])]],disp="p",dispminimum=NA,limits=limits[[i]],title=names(results)[j],letter=LETTERS[j],legend=ifelse(j==1,T,F))
			}
			dev.off()
		}

		# plot PCoAs of optimal solutions, using distance matrix from interval density function
		results=lapply(results,function(a) a[[length(a)]])		
		res1=lapply(results,transformObjectives)
		bl=best.layout(6,2)
		openDevice(output,name="solutions_ord",wid=dim(bl)[2]*3.5,hei=dim(bl)[1]*3.5)
		layout(bl)
		for(i in 2:6) {		# each restriction scenario (except the no-restriction because it's a huge matrix)
			ssr=list()
			subsetted=list()
			for(j in 1:length(res1)) {
				limits=setLimits(results[[j]])
				ssr[[j]]=subset(res1[[j]],limits[[i]])
				subsetted[[j]]=ssr[[j]]$subsetted
			}
			solutions=ordinateSolutions(subsetted,title=names(limits)[i],letter=LETTERS[i-1],nominalset=baseindex,xlim=c(-0.4,0.4),ylim=c(-0.4,0.4))	#,xlim=c(-0.62,0.5)
		}
		# draw legend
		plot.new()
		plot.window(xlim=c(0,1),ylim=c(0,1))
		pt.lwd=rep(0.5,length(res1))
		pch=rep(21,length(res1))
		pt.cex=rep(1.25,length(res1))
		leg=names(res1)
		pt.lwd[baseindex]=1.5
		pch[baseindex]=24
		pt.cex[baseindex]=1.5
		leg[baseindex]=paste(leg[baseindex],"(base simulation)")
		
		legend("center",leg=leg,col="black",pt.lwd=pt.lwd,pt.bg=1:length(res1),pch=pch,pt.cex=pt.cex,bty="n",cex=1.2,title=title)
		dev.off()
	},{
		files=dir("Results/",pattern="optim_")
		file=paste("Results/",files[menu(files)],sep="")
		load(file)
		if(!exists("results")) results=res
		if(length(results)<2) stop("This file has only one optimization run.")

		for(i in 1:NOBJ) {
			bl=best.layout(length(results),2)
			openDevice(output,name=paste("levelplots_obj",i,sep=""),wid=dim(bl)[2]*3.5,hei=dim(bl)[1]*3.5)
			layout(bl)
			for(j in 1:length(results))
		   		visualize(results[[j]][[length(results[[j]])]],disp="o",color=T,dispminimum=NA,ranges=list(NULL,NULL,NULL,c(-5,5),NULL,NULL),title=names(results)[j],letter=LETTERS[j],objectives=i)
		   	dev.off()
		}
	})
},{
	if(!require(mcomp)) stop("You must compile and install the modified package in order to run this option.")
	compromise.solutions=list( c(72,100,3,0.29,0.15), c(75,100,3,0.38,0.08), c(76,100,6,0.31,0.10) )
	choices=c(sapply(compromise.solutions,format.pars),"Write a solution manually")
	cat("\nChoose one or more solutions to simulate, separated by commas:\n\n")
	cat(paste(1:length(choices),choices,sep=": "),sep="\n")
	cat("\nSelection(s): ")
	choice=as.numeric(strsplit(readline(),",",fixed=T)[[1]])
	if(length(choices)>1 && any(choice==length(choices))) {cat("Cannot mix the last option with others.\n\n");choice=length(choices)}
        
    if(length(choice)==1 && choice==length(choices)) {
		cat("Enter solutions to simulate, in the form:\nMan.Inter.Gr.1, ..., Man.Inter.Gr.N, Prop.Gr.1, ..., Prop.Gr.N-1\nWrite each solution in a separate line. Terminate with a blank line.\n\nExample:\n12,27,50,0.3,0.45\nthis would represent a solution with 3 groups:\n30% managing every 12 years, 45% every 27 yr and 25% every 50 yr.\n")
		tmp=read.csv(stdin(),h=F)
		pars=list()
		for(i in 1:dim(tmp)[1]) {
		    pars=c(pars,list(tmp[i,][!is.na(tmp[i,])]))
		}
    } else {
    	pars=compromise.solutions[choice]
    }

	cat("Warning: this consumes huge amounts of memory. Press ENTER to start.")
	readline()
	otherpars$n.landscape.repetitions=100L
	#otherpars$n.years=500L
	
	schedule=list(
		list(
			task="examine",
			uncertainties=c(100),
			increasing.uncertainty=1,
			solutions=pars,
			landscape.files="landscapes_20_100.bin",
			legend.func=function() {
				legend("top",leg=sapply(pars,format.pars),col=1:3,lty=1,lwd=2,title="Solutions",bty="n",cex=1.1)
			}
		
		)
	)
	res=processTasks(schedule)
},{
	files=dir("Results/",pattern="optim_")
	file=paste("Results/",files[menu(files)],sep="")
	load(file)
	if(!exists("results")) results=res

	if(length(results)>1) {
		cat("There are",length(results),"runs in this file. Which one do you want?\n")
		if(is.null(names(results)))
			run=menu(1:length(results))
		else
			run=menu(names(results))
		results=results[[run]]
	}
	
	dh=numeric(length(results))
	# reference point for hypervolume indicator calculation - it's the maximum of each objective across all runs
	referencepoint=apply(sapply(results,function(a) {apply(a$val,2,max)}),1,max)
	for(i in 1:length(results)) {
		dh[i]=dominatedHypervolume(results[[i]],referencepoint)
		cat("Dominated hypervolume:",dh[i],"\n")
		flush.console()
		#visualize(results,disp="p",limits=limits,dispminimum=NA)
	}
	conv=matrix(c(otherpars$gener*1,dh),nc=2,dimnames=list(NULL,c("Generation","Hypervolume indicator")))
	openDevice(output,name="convergence",wid=1*3.5,hei=1*3.5)
	par(mai=c(0.35,0.35,0.2,0.1),mgp=c(1.1,0.3,0),tcl=-0.3,cex=0.8)
	plot(conv,type="l",lwd=1.5,bty="l")
	points(conv,pch=19)
	dev.off()
},{
		openDevice(output,name="uncertainty_distribution",wid=1*3.5,hei=1*3.5)
		plotUncertainty(get.interval.gamma(100,1,num=10000))
		dev.off()
})


